@@ -401,8 +401,7 @@ void InMemoryView::processPath(
     return;
   }
 
-  if (w_string_equal(pending.path, rootPath_) ||
-      (pending.flags & W_PENDING_CRAWL_ONLY)) {
+  if (pending.path == rootPath_ || (pending.flags & W_PENDING_CRAWL_ONLY)) {
     crawler(root, view, coll, pending, pendingCookies);
   } else {
     statPath(*root, root->cookies, view, coll, pending, pre_stat);
@@ -468,7 +467,7 @@ void InMemoryView::crawler(
           pending.now,
           "getFileInformation",
           err.code());
-      view.markDirDeleted(*watcher_, dir, getClock(pending.now), true);
+      view.markDirDeleted(dir, getClock(pending.now), true);
       return;
     }
   }
@@ -494,7 +493,7 @@ void InMemoryView::crawler(
     logf(DBG, "startWatchDir({}) threw {}\n", path, err.what());
     handle_open_errno(
         *root, dir->getFullPath(), pending.now, "opendir", err.code());
-    view.markDirDeleted(*watcher_, dir, getClock(pending.now), true);
+    view.markDirDeleted(dir, getClock(pending.now), true);
     return;
   }
 
@@ -677,7 +676,12 @@ void InMemoryView::crawlerParallel(
   std::shared_ptr<CrawlerFileSystem> fs =
       std::make_shared<CrawlerFileSystem>(fileSystem_, root, watcher_);
   size_t threadCountHint = config_.getInt("parallel_crawl_thread_count", 0);
-  ParallelWalker walker{std::move(fs), path, threadCountHint};
+  ParallelWalker walker{
+      std::move(fs),
+      path,
+      root->allow_crawling_other_mounts ? std::nullopt
+                                        : std::optional{root->stat},
+      threadCountHint};
 
   // Step 1: Process readDir results.
   while (true) {
@@ -811,7 +815,7 @@ bool did_file_change(
 } // namespace
 
 void InMemoryView::statPath(
-    const RootConfig& root,
+    const Root& root,
     const CookieSync& cookies,
     ViewDatabase& view,
     PendingChanges& coll,
@@ -830,10 +834,10 @@ void InMemoryView::statPath(
   }
 
   auto& path = pending.path;
-  w_check(path, "must have path");
+  w_check(!path.empty(), "must have path");
   auto dir_name = pending.path.dirName();
   auto file_name = pending.path.baseName();
-  w_check(dir_name, "must have dir_name");
+  w_check(!dir_name.empty(), "must have dir_name");
   auto parentDir = view.resolveDir(dir_name, true);
 
   auto file = parentDir->getChildFile(file_name);
@@ -887,7 +891,7 @@ void InMemoryView::statPath(
       errcode == error_code::not_a_directory) {
     /* it's not there, update our state */
     if (dir_ent) {
-      view.markDirDeleted(*watcher_, dir_ent, getClock(pending.now), true);
+      view.markDirDeleted(dir_ent, getClock(pending.now), true);
       log(DBG,
           "getFileInformation(",
           path,
@@ -906,7 +910,7 @@ void InMemoryView::statPath(
             file->getName(),
             " deleted\n");
         file->exists = false;
-        view.markFileChanged(*watcher_, file, getClock(pending.now));
+        view.markFileChanged(file, getClock(pending.now));
       }
     } else {
       // It was created and removed before we could ever observe it
@@ -914,7 +918,7 @@ void InMemoryView::statPath(
       // representation of it now, so that subscription clients can
       // be notified of this event
       file = view.getOrCreateChildFile(
-          *watcher_, parentDir, file_name, getClock(pending.now));
+          parentDir, file_name, getClock(pending.now));
       log(DBG,
           "getFileInformation(",
           path,
@@ -923,7 +927,7 @@ void InMemoryView::statPath(
           " and file node was NULL. "
           "Generating a deleted node.\n");
       file->exists = false;
-      view.markFileChanged(*watcher_, file, getClock(pending.now));
+      view.markFileChanged(file, getClock(pending.now));
     }
 
     if (!viaPwalk &&
@@ -958,7 +962,7 @@ void InMemoryView::statPath(
   } else {
     if (!file) {
       file = view.getOrCreateChildFile(
-          *watcher_, parentDir, file_name, getClock(pending.now));
+          parentDir, file_name, getClock(pending.now));
     }
 
     if (!file->exists) {
@@ -981,7 +985,7 @@ void InMemoryView::statPath(
           st.size,
           path);
       file->exists = true;
-      view.markFileChanged(*watcher_, file, getClock(pending.now));
+      view.markFileChanged(file, getClock(pending.now));
 
       // If the inode number changed then we definitely need to recursively
       // examine any children because we cannot assume that the kernel will
@@ -993,6 +997,7 @@ void InMemoryView::statPath(
     }
 
     memcpy(&file->stat, &st, sizeof(file->stat));
+    watcher_->startWatchFile(file);
 
     if (st.isDir()) {
       if (dir_ent == NULL) {
@@ -1002,6 +1007,15 @@ void InMemoryView::statPath(
         dir_ent->last_check_existed = true;
       }
 
+      if (!(root.allow_crawling_other_mounts ||
+            isOnSameMount(root.stat, st, path.c_str()))) {
+        logf(
+            DBG,
+            "Ignoring {} as it doesn't reside on the same mount point as the root directory\n",
+            pending.path);
+        return;
+      }
+
       // Don't recurse if our parent is an ignore dir or via crawlerParallel
       // (already recursive)
       if (!viaPwalk &&
@@ -1041,7 +1055,7 @@ void InMemoryView::statPath(
     } else if (dir_ent) {
       // We transitioned from dir to file (see fishy.php), so we should prune
       // our former tree here
-      view.markDirDeleted(*watcher_, dir_ent, getClock(pending.now), true);
+      view.markDirDeleted(dir_ent, getClock(pending.now), true);
     }
   }
 }
